"""
Translate an options data structure into command line args
"""
import sys
from typing import Any
from typing import Dict


class DefaultArg:
    ...


def __init__(hub):
    hub.args.DEFAULT = DefaultArg()
    hub.args.PARSER = None
    hub.args.SUBPARSER = None
    hub.args.SUBPARSERS = {}


def keys(hub, opts):
    """
    Return the keys in the right order
    """
    return sorted(opts, key=lambda k: (opts[k].get("display_priority", sys.maxsize), k))


def setup(hub, raw_cli: Dict[str, Any]) -> Dict[str, Any]:
    """
    Take in a pre-defined dict and translate it to args

    opts dict:
        <arg>:
            [group]: foo
            [default]: bar
            [action]: store_true
            [options]: # arg will be turned into --arg
              - '-A'
              - '--cheese'
            [choices]:
              - foo
              - bar
              - baz
            [nargs]: +
            [type]: int
            [dest]: cheese
            help: Some great help message
    """
    parser_definition = []
    for arg in hub.args.init.keys(raw_cli):
        if arg in ("_argparser_",):
            continue

        # Pass the raw_cli and args through all the comps plugins
        ret = hub.comps.init.get(raw_cli, arg)
        parser_definition.append(ret)

    # Apply this config to the root parser
    hub.args.parser.apply(hub.args.PARSER, parser_definition)



"""
The os sub is used to gather configuration options from the OS facility
to send configuration options into applications.
"""
import itertools
import os
from typing import Dict


def gather(hub, raw: Dict) -> Dict[str, Dict]:
    """
    Collect the keys that need to be found and pass them to the
    os specific loaded plugin
    """
    ret = {}
    for sec, imp in itertools.product(hub.config.CONFIG_SECTIONS, raw):
        for key in raw[imp].get(sec, ()):
            osvar = raw[imp][sec][key].get("os")
            if osvar is None:
                continue
            val = os.environ.get(osvar.upper())
            if val is None:
                continue
            src = raw[imp][sec][key].get("source", imp)
            if src not in ret:
                ret[src] = {}
            ret[src][key] = val
    return ret



"""
Gather groups for a parser or subparser
"""
from typing import Any
from typing import Dict


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    comps = raw_cli[arg]

    if comps.get("positional"):
        # The action will define the dest for positional args
        return {}

    args = [f"--{arg.replace('_', '-')}"]
    for o_str in comps.get("options", ()):
        if len(o_str) == 1:
            o_str = f"-{o_str}"
        elif not o_str.startswith("-"):
            o_str = f"--{o_str}"
        if o_str not in args:
            args.append(o_str)

    return {"parser_args": args}



"""
Gather actions for a parser or subparser
"""
import inspect
from typing import Any
from typing import Dict


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    comps = raw_cli[arg]
    kwargs = {}
    defaults = {}

    kwargs["action"] = action = comps.get("action", None)

    if action is None:
        # Non existing option defaults to a StoreAction in argparse
        action = hub.args.PARSER._registry_get("action", None)

    if isinstance(action, str):
        signature = inspect.signature(
            hub.args.PARSER._registry_get("action", action).__init__
        )
    else:
        signature = inspect.signature(action.__init__)

    if "dest" in signature.parameters:
        # This is the default for positional args
        kwargs["dest"] = comps.pop("dest", arg)
    if "help" in signature.parameters:
        kwargs["help"] = comps.pop("help", "THIS NEEDS SOME DOCUMENTATION!!")

    for comp in comps:
        if comp in signature.parameters:
            kwargs[comp] = comps[comp]

    return {"parser_kwargs": kwargs, "defaults": defaults}



import sys


try:
    import aiologger
    import aiologger.handlers.streams
    import aiologger.handlers.files
    import aiologger.handlers.base
    import aiologger.levels
    import aiologger.formatters.base
    import aiologger.records

    HAS_AIOLOGGER = (True,)
except ImportError as e:
    HAS_AIOLOGGER = False, str(e)


def __init__(hub):
    hub.log.LOGGER = {}
    hub.log.FILE_HANDLER = None
    hub.log.STREAM_HANDLER = None


def __virtual__(hub):
    return HAS_AIOLOGGER


def _get_logger(hub, name: str = ""):
    if name not in hub.log.LOGGER:
        hub.log.LOGGER[name]: aiologger.Logger = aiologger.Logger(
            name=name, loop=hub.pop.Loop
        )
        hub.log.LOGGER[name].level = hub.log.INT_LEVEL
        if hub.log.FILE_HANDLER:
            hub.log.LOGGER[name].handlers.append(hub.log.FILE_HANDLER)
        if hub.log.STREAM_HANDLER:
            hub.log.LOGGER[name].handlers.append(hub.log.STREAM_HANDLER)
    return hub.log.LOGGER[name]


def setup(hub, conf):
    """
    Given the configuration data set up the logger
    """
    # Use the saved root logger
    root = _get_logger(hub, name="")

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.level = hub.log.INT_LEVEL
    cf = aiologger.formatters.base.Formatter(
        fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"]
    )
    ch = aiologger.handlers.streams.AsyncStreamHandler(
        formatter=cf, loop=hub.pop.Loop, stream=sys.stderr
    )
    ch._level = hub.log.INT_LEVEL
    root.add_handler(ch)
    hub.log.STREAM_HANDLER = ch

    ff = aiologger.formatters.base.Formatter(
        fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"]
    )
    _, kwargs = hub.render.cli.args(conf["log_handler_options"])
    fh = aiologger.handlers.files.AsyncTimedRotatingFileHandler(
        conf["log_file"], **kwargs
    )
    fh._level = hub.log.INT_LEVEL
    fh.formatter = ff
    root.add_handler(fh)
    hub.log.FILE_HANDLER = fh

    # Put all these functions higher up on the hub
    hub.log.log = hub.log["async"].log
    hub.log.trace = lambda msg, *args, **kwargs: hub.log.log(
        level=5, msg=msg, *args, **kwargs
    )
    hub.log.debug = lambda msg, *args, **kwargs: hub.log.log(
        level=aiologger.levels.LogLevel.DEBUG, msg=msg, *args, **kwargs
    )
    hub.log.info = lambda msg, *args, **kwargs: hub.log.log(
        level=aiologger.levels.LogLevel.INFO, msg=msg, *args, **kwargs
    )
    hub.log.warning = lambda msg, *args, **kwargs: hub.log.log(
        level=aiologger.levels.LogLevel.WARNING, msg=msg, *args, **kwargs
    )
    hub.log.error = lambda msg, *args, **kwargs: hub.log.log(
        level=aiologger.levels.LogLevel.ERROR, msg=msg, *args, **kwargs
    )
    hub.log.critical = lambda msg, *args, **kwargs: hub.log.log(
        level=aiologger.levels.LogLevel.CRITICAL, msg=msg, *args, **kwargs
    )



"""
Define the yaml loader interface
"""

try:
    import toml

    HAS_TOML = (True,)
except ImportError as e:
    HAS_TOML = False, str(e)

__virtualname__ = "toml"


def __virtual__(hub):
    return HAS_TOML


def load(hub, path):
    """
    use toml to read in a file
    """
    try:
        with open(path) as fp_:
            return toml.loads(fp_.read())
    except FileNotFoundError:
        pass
    return {}


def render(hub, val):
    """
    Take the string and render it in json
    """
    return toml.loads(val)



"""
Define the yaml loader interface
"""

try:
    import yaml

    HAS_YAML = (True,)
except ImportError as e:
    HAS_YAML = False, str(e)

__virtualname__ = "yaml"


def __virtual__(hub):
    return HAS_YAML


def load(hub, path):
    """
    use yaml to read in a file
    """
    try:
        with open(path, "rb") as fp_:
            return yaml.safe_load(fp_.read())
    except FileNotFoundError:
        pass
    return {}


def render(hub, val):
    """
    Take the string and render it in json
    """
    return yaml.safe_load(val)



from typing import Dict


def apply(
    hub,
    raw: Dict,
    raw_cli: Dict,
    cli: Dict,
    cli_args: Dict,
    os_vars: Dict,
    configs: Dict,
) -> Dict:
    """
    This assumes that we are using the namespace approach;
    This makes the config structure the easiest, meaning that components are namespaced by the user.
    Some other additional pattern could be added and this chunk could be made pluggable.
    Order of config loading:
    # Defaults (raw)
    # Config files (configs)
    # OS (os_vars)
    # CLI (cli_args)
    """
    ret = {}
    ret = hub.config.order.defaults(raw, ret)
    ret = hub.config.order.config(configs, ret)
    ret = hub.config.order.os_vars(os_vars, ret)
    ret = hub.config.order.cli_vars(raw_cli, cli_args, cli, ret)
    return ret


def defaults(hub, raw: Dict, ret: Dict) -> Dict:
    """
    Get the CONFIG values with defaults
    """
    for imp in raw:
        ret[imp] = {}
        for key, data in raw[imp]["CONFIG"].items():
            if "default" in data:
                ret[imp][key] = data["default"]
    return ret


def config(hub, configs, ret) -> Dict:
    """
    Get values from the config file
    """
    for imp in configs:
        if imp not in ret:
            ret[imp] = {}
        for key in configs[imp]:
            ret[imp][key] = configs[imp][key]
    return ret


def os_vars(hub, os_vars: Dict[str, Dict], ret: Dict[str, Dict]) -> Dict:
    """
    Read OS variables
    """
    for imp in os_vars:
        for key in os_vars[imp]:
            ret[imp][key] = os_vars[imp][key]
    return ret


def cli_vars(
    hub,
    raw_cli: Dict[str, Dict],
    cli_args: Dict[str, Dict],
    cli: Dict[str, Dict],
    ret: Dict,
) -> Dict:
    """
    Read CLI variables
    """
    for key in cli_args:
        if key in raw_cli:
            if "source" in raw_cli[key]:
                ret[raw_cli[key]["source"]][key] = cli_args[key]
            else:
                ret[cli][key] = cli_args[key]
    return ret



import os
import sys

CLI_CONFIG = {
    "log_file": {"dyne": "__cli__", "os": "POP_LOG_FILE", "subcommands": ["_global_"]},
    "log_level": {
        "dyne": "__cli__",
        "os": "POP_LOG_LEVEL",
        "subcommands": ["_global_"],
    },
    "log_fmt_logfile": {
        "dyne": "__cli__",
        "os": "POP_LOG_FMT_LOGFILE",
        "subcommands": ["_global_"],
    },
    "log_fmt_console": {
        "dyne": "__cli__",
        "os": "POP_LOG_FMT_CONSOLE",
        "subcommands": ["_global_"],
    },
    "log_datefmt": {
        "dyne": "__cli__",
        "os": "POP_LOG_DATEFMT",
        "subcommands": ["_global_"],
    },
    "log_plugin": {
        "dyne": "__cli__",
        "os": "POP_LOG_PLUGIN",
        "subcommands": ["_global_"],
    },
    "log_handler_options": {
        "dyne": "__cli__",
        "subcommands": ["_global_"],
        "nargs": "*",
    },
    "version": {
        "dyne": "__cli__",
        "default": False,
        "action": "store_true",
        "help": "Display version information",
    },
    "config_template": {
        "dyne": "__cli__",
        "subcommands": ["_global_"],
        "default": False,
        "action": "store_true",
        "help": "Output a config template for this command",
    },
    "versions_report": {
        "dyne": "__cli__",
        "subcommands": ["_global_"],
        "default": False,
        "action": "store_true",
        "help": "Output a version report for reporting bugs",
    },
}

CONFIG = {
    "log_file": {
        "dyne": "__cli__",
        "default": f"{os.path.splitext(os.path.split(sys.argv[0])[1])[0]}.log",
        "help": "The location of the log file",
        "group": "Logging Options",
    },
    "log_level": {
        "dyne": "__cli__",
        "default": "warning",
        "help": "Set the log level, either quiet, info, warning, debug or error",
        "group": "Logging Options",
    },
    "log_fmt_logfile": {
        "dyne": "__cli__",
        "default": "%(asctime)s,%(msecs)03d [%(name)-17s][%(levelname)-8s] %(message)s",
        "help": "The format to be given to log file messages",
        "group": "Logging Options",
    },
    "log_fmt_console": {
        "dyne": "__cli__",
        "default": "[%(levelname)-8s] %(message)s",
        "help": "The log formatting used in the console",
        "group": "Logging Options",
    },
    "log_datefmt": {
        "dyne": "__cli__",
        "default": "%H:%M:%S",
        "help": "The date format to display in the logs",
        "group": "Logging Options",
    },
    "log_plugin": {
        "dyne": "__cli__",
        "default": "basic",
        "help": "The logging plugin to use",
        "group": "Logging Options",
    },
    "log_handler_options": {
        "dyne": "__cli__",
        "default": [],
        "help": "kwargs that should be passed to the logging handler used by the log_plugin",
        "group": "Logging Options",
    },
}
DYNE = {
    "args": ["args"],
    "comps": ["comps"],
    "config": ["config"],
    "log": ["log"],
    "render": ["render"],
}



import logging.handlers


def setup(hub, conf):
    """
    Given the configuration data set up the logger.
    """
    root = logging.getLogger()

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)



from typing import Dict
from typing import List
from typing import Tuple

try:
    import yaml

    HAS_YAML = (True,)
except ImportError as e:
    HAS_YAML = False, str(e)

__virtualname__ = "cli"


def __virtual__(hub):
    return HAS_YAML


def load(hub, path):
    try:
        with open(path, "rb") as fp_:
            ret = {}
            for line in fp_:
                ret.update(hub.render.cli.render(line))
            return ret
    except FileNotFoundError:
        pass
    return {}


def render(hub, val: List[str] or str) -> List[str]:
    """
    Take the string and render it in json
    """
    ret = []
    if isinstance(val, str):
        val = [val]
    for v in val:
        if "=" in v:
            key, v = v.split("=", maxsplit=1)
            ret.append({key: yaml.safe_load(v)})
        else:
            ret.append(yaml.safe_load(v))

    return ret


def args(hub, val: List[str] or str) -> Tuple[List[str], Dict[str, str]]:
    """
    Take a string and convert it to args and kwargs
    """
    args = []
    kwargs = {}
    for v in hub.render.cli.render(val):
        if isinstance(v, dict):
            kwargs.update(v)
        else:
            args.append(v)

    return args, kwargs



from typing import Any
from typing import Dict

GLOBAL_SUBCOMMAND = "_global_"


def init(hub):
    hub.args.parser.init()
    if hub.args.SUBPARSER is None:
        hub.args.SUBPARSER = hub.args.PARSER.add_subparsers(dest="_subparser_")


def add(hub, arg: str, **kwargs):
    hub.args.subcommands.init()
    hub.args.SUBPARSERS[arg] = hub.args.SUBPARSER.add_parser(arg, **kwargs)


def parse(hub, raw: Dict[str, Any], cli: str):
    """
    Look over the data and extract and set up the subparsers for subcommands
    """
    subs = raw.get(cli, {}).get("SUBCOMMANDS")
    if not subs:
        return

    for arg in hub.args.init.keys(subs):
        if arg in ("_argparser_",):
            continue
        comps = subs[arg]
        kwargs = {}
        if "help" in comps:
            kwargs["help"] = comps["help"]
        if "desc" in comps:
            kwargs["description"] = comps["desc"]
        hub.args.subcommands.add(arg, **kwargs)


def apply(
    hub,
    root_parser,
    context: Dict[str, Any],
    definition: Dict[str, Any],
):
    for subcommand in definition.pop("subcommands"):
        if subcommand == GLOBAL_SUBCOMMAND:
            # Add it to the main parser first
            hub.args.parser.apply(
                parser=root_parser,
                context=context,
                arg_definitions=[definition],
            )

            # Add it to all subcommands
            for name in hub.args.SUBPARSERS:
                hub.args.subcommands.apply_single(
                    subparser=name, definition=definition, context=context
                )
        else:
            hub.args.subcommands.apply_single(
                subparser=subcommand, definition=definition, context=context
            )


def apply_single(hub, subparser: str, context, definition):
    if subparser not in hub.args.SUBPARSERS:
        return
    if subparser not in context["subparser_contexts"]:
        context["subparser_contexts"][subparser] = hub.args.parser.context()
    hub.args.parser.apply(
        parser=hub.args.SUBPARSERS[subparser],
        arg_definitions=[definition],
        context=context["subparser_contexts"][subparser],
    )



"""
Support embedding version number lookup into cli
"""
import sys


def run(hub, opts):
    """
    Output a config file based on the current command
    """
    hub.pop.sub.add(dyne_name="rend")
    outputter = opts.get("rend", {}).get("output") or "yaml"

    output_opts = hub.config.template.get(opts)

    out = hub.output[outputter].display(output_opts)

    print(out)
    sys.exit(0)


def get(hub, opts) -> dict:
    output_opts = {}
    # Trim old values from opt
    for dyne, config in opts.items():
        if not config:
            continue

        output_config = {}
        for k, v in config.items():
            # config_template will be "True" for this run, which would make the config file always result in exit
            if k in ("config_template", "config", "versions_report"):
                continue
            output_config[k] = v
        output_opts[dyne] = output_config
    return output_opts



"""
The render process for the cli does NOT use the rend project for a couple
of reasons
1. Config needs to be very early in the startup, therefore it cannot use
   asyncio, all rend funcs use asyncio
2. Config will not allow for template wrapping as the render is just a
   single command line render
"""
from typing import Any
from typing import List


def process(hub, renderer: str, value: str):
    """
    Take a renderer and a value, process it, and return the processed value

    This is intended to load a string through the config render system
    """
    return hub.render[renderer].render(value)


def pipe(hub, dpipe: List[str], data: Any):
    """
    Given a render pipe, render the given data
    """
    for render in dpipe:
        if isinstance(render, bytes):
            render = render.decode()
        data = hub.render.init.process(render, data)
    return data


def load_file(hub, renderer: str, file_name: str):
    """
    Load up a file with the passed renderer unless the file contains a
    renderer she-bang line
    """
    with open(file_name, "rb") as rfh:
        data = rfh.read()
    if not data:
        return {}
    if data.startswith(b"#!"):
        dpipe = data[2 : data.index(b"\n")].split(b"|")
        return hub.render.init.pipe(dpipe, data)
    else:
        return hub.render.init.process(renderer, data)



def __init__(hub):
    """
    Load the subdirs for conf
    """
    for dyne in ("args", "comps", "log", "render"):
        hub.pop.sub.add(dyne_name=dyne)
    hub.config.ARGS = {}
    hub.config.SECTIONS = ("CONFIG", "CLI_CONFIG", "SUBCOMMANDS", "DYNE")
    hub.config.CONFIG_SECTIONS = ("CONFIG", "CLI_CONFIG")



"""
This sub is used to set up logging for pop projects and injects logging
options into conf making it easy to add robust logging
"""
import logging


def __init__(hub):
    """
    Set up variables used by the log subsystem
    """
    logging.addLevelName(5, "TRACE")
    hub.log.LEVEL = {
        "notset": logging.NOTSET,
        "trace": 5,
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARN,
        "warning": logging.WARNING,
        "error": logging.ERROR,
        "fatal": logging.FATAL,
        "critical": logging.CRITICAL,
    }
    log = logging.getLogger(__name__)

    # These should be overwritten by the integrated logger, but here's a contingency
    hub.log.INT_LEVEL = log.getEffectiveLevel()
    hub.log.log = log.log
    hub.log.trace = lambda msg, *args, **kwargs: log.log(5, msg, *args, **kwargs)
    hub.log.debug = log.debug
    hub.log.info = log.info
    hub.log.critical = log.critical
    hub.log.warning = log.warning
    hub.log.error = log.error



"""
Find the conf.py files specified in sources
"""
import copy
import importlib
import os.path
from typing import Dict
from typing import List
from typing import Tuple

import dict_tools.update


def load(hub, sources: List[str], dyne_names: List[str], cli: str):
    """
    Look over the sources list and find the correct conf.py files
    """
    # Dynamic names
    # First gather the defined sources, they are the authoritative ones
    # Then detect what the dynamic names are in the source
    # Merged the sources dyne names with any passed dyne names
    # Load up and extend the raw with all of the dynamic names
    dyne = hub.pop.dyne.get()
    if not isinstance(sources, list):
        sources = [sources]
    raw = hub.config.dirs.find_configs(sources)
    # Makes sure everything is unique
    dnames = set(dyne_names)
    for source in sources:
        dnames.update(raw[source]["DYNE"].keys())
    for name in dnames:
        if name in dyne:
            if name not in raw:
                raw[name] = {"CONFIG": {}, "CLI_CONFIG": {}}
            dyne_data = dyne[name]
            if "CONFIG" in dyne_data:
                dyne_data, config_draw = hub.config.dirs.parse_config(dyne_data, cli)
                dict_tools.update.update(raw[cli]["CONFIG"], config_draw)
            if "CLI_CONFIG" in dyne_data:
                cli_draw = hub.config.dirs.parse_cli(dyne_data, cli)
                dict_tools.update.update(raw[cli]["CLI_CONFIG"], cli_draw)
            if "SUBCOMMANDS" in dyne_data:
                subcmd_draw = hub.config.dirs.parse_subcommand(dyne_data, cli)
                dict_tools.update.update(raw[cli]["SUBCOMMANDS"], subcmd_draw)
    return raw


def find_configs(hub, sources: List[str]):
    raw = {}
    for source in sources:
        try:
            path, data = hub.config.dirs.import_conf(source)
        except ImportError:
            continue
        dict_tools.update.update(raw, data)
    return raw


def import_conf(hub, imp: str) -> Tuple[str, Dict]:
    """
    Load up a python path, parse it and return the conf dataset
    """
    ret = {imp: {}}
    cmod = importlib.import_module(f"{imp}.conf")
    path = os.path.dirname(cmod.__file__)
    for section in hub.config.SECTIONS:
        ret[imp][section] = copy.deepcopy(getattr(cmod, section, {}))
    return path, ret


def parse_config(hub, dyne_data: Dict[str, Dict], cli: str) -> Tuple[Dict, Dict]:
    config_draw = {}
    for key, val in dyne_data["CONFIG"].items():
        new_dyne = val.get("dyne")
        if new_dyne == "__cli__":
            new_dyne = cli
        if new_dyne:
            val["source"] = new_dyne
            config_draw[key] = val
            if (
                key in dyne_data.get("CLI_CONFIG", {})
                and "dyne" not in dyne_data["CLI_CONFIG"][key]
            ):
                dyne_data["CLI_CONFIG"][key]["dyne"] = new_dyne
    return dyne_data, config_draw


def parse_cli(hub, dyne_data: Dict, cli: str) -> Dict:
    cli_draw = {}
    for key, val in dyne_data["CLI_CONFIG"].items():
        new_dyne = val.get("dyne")
        if new_dyne == "__cli__":
            new_dyne = cli
        if new_dyne:
            val["source"] = new_dyne
            cli_draw[key] = val
    return cli_draw


def parse_subcommand(hub, dyne_data: Dict, cli: str) -> Dict:
    subcmd_draw = {}
    for key, val in dyne_data["SUBCOMMANDS"].items():
        new_dyne = val.get("dyne")
        if new_dyne == "__cli__":
            new_dyne = cli
        if new_dyne:
            val["source"] = new_dyne
        if new_dyne == cli:
            subcmd_draw[key] = val
    return subcmd_draw


def verify(hub, opts):
    """
    Verify that the environment and all named directories in the
    configuration exist
    """
    for imp in opts:
        for key in opts[imp]:
            if key == "root_dir":
                continue
            if key == "config_dir":
                continue
            if key.endswith("_dir"):
                if not os.path.isdir(opts[imp][key]):
                    os.makedirs(opts[imp][key])



"""
Configuration file core loading functions
"""
import pathlib

import dict_tools.update


def parse(hub, raw, cli, os_vars, cli_args, loader):
    """
    Determine if a config file or a config dir has been set up and load it up!
    """
    # This function is the entry point for the config.conf_file sub
    # Figure out what config file value to use in this priority
    default = raw[cli]["CONFIG"].get("config", {}).get("default")
    default_dir = raw[cli]["CONFIG"].get("config_dir", {}).get("default")
    if cli in os_vars:
        os_conf = os_vars[cli].get("config", default)
        os_dir = os_vars[cli].get("config_dir", default_dir)
    else:
        os_conf = os_vars.get("config", default)
        os_dir = os_vars.get("config_dir", default_dir)
    conf = cli_args.get("config", os_conf)
    dir_ = cli_args.get("config_dir", os_dir)
    file_opts = {}
    dir_opts = {}
    if dir_:
        file_opts = hub.config.file.load_dir(dir_, loader)
    if conf:
        dir_opts = hub.config.file.load(conf, loader)
    ret = dict_tools.update.update(dir_opts, file_opts)
    return ret


def load(hub, paths, loader, includes=True):
    """
    Load a single configuration file
    """
    opts = {}
    if not isinstance(paths, list):
        paths = [paths]
    add = []
    for fn in paths:
        for path in pathlib.Path(fn).rglob("**"):
            add.append(str(path))
    paths.extend(add)
    for fn in paths:
        fn_data = hub.render.init.load_file(loader, fn)
        if includes:
            fn_data = hub.config.file.proc_include(fn, fn_data, loader)
        dict_tools.update.update(opts, fn_data)
    return opts


def load_dir(
    hub,
    confdir,
    loader,
    includes=True,
    recurse=True,
):
    """
    Load takes a directory location to scan for configuration files. These
    files will be read in.
    """
    opts = {}
    if not isinstance(confdir, list):
        confdir = [confdir]
    confdirs = []
    for dirs in confdir:
        if not isinstance(dirs, (list, tuple)):
            dirs = [dirs]
        for dir_ in dirs:
            for g in pathlib.Path(dir_).rglob("**"):
                confdirs.append(g)
    paths = []
    for dir_ in confdirs:
        dirpaths = []
        if not dir_.is_dir():
            continue
        if not recurse:
            for path in dir_.iterdir():
                if path.is_dir():
                    # Don't process directories
                    continue
                dirpaths.append(str(path))
        else:
            for p in dir_.iterdir():
                dirpaths.append(str(p))

        # Sort confdir directory paths like:
        # /b.txt
        # /c.txt
        # /a/x.txt
        # /b/x.txt
        paths.extend(sorted(dirpaths, key=lambda p: (p.count(pathlib.os.path.sep), p)))
    dict_tools.update.update(opts, hub.config.file.load(paths, loader, includes))
    return opts


def proc_include(hub, fn, opts, loader):
    """
    Process include and include_dir
    """
    dirname = pathlib.Path(fn).parent
    root = pathlib.Path(pathlib.os.sep).absolute()
    if opts.get("include_dir"):
        idir = pathlib.Path(opts.pop("include_dir"))
        pathlib.sep
        if not str(idir).startswith(str(root)):
            idir = dirname / idir
        dict_tools.update.update(opts, hub.config.file.load_dir(str(idir), loader))
        hub.config.file.proc_include(str(idir / "f"), opts, loader)
    if opts.get("include"):
        ifn = opts.pop("include")
        if not ifn.startswith(str(root)):
            ifn = dirname / ifn
        dict_tools.update.update(opts, hub.config.file.load(ifn, loader))
        hub.config.file.proc_include(ifn, opts, loader)
    return opts



from typing import List


def load(
    hub,
    sources: List[str],
    cli: str = None,
    dyne_names: List[str] = None,
    loader: str = "yaml",
    parse_cli: bool = True,
    logs: bool = True,
):
    """
    Load up the configs from the integrate system
    """
    if not isinstance(sources, list):
        sources = [sources]
    sources.append("pop_config")
    if dyne_names is None:
        dyne_names = []
    raw = hub.config.dirs.load(sources, dyne_names, cli)
    os_vars = hub.config.os_vars.gather(raw)
    cli_args, raw_cli = hub.args.cli.gather(raw, cli, parse_cli)
    if cli_args.get("version"):
        hub.config.version.run(cli)
    configs = hub.config.file.parse(raw, cli, os_vars, cli_args, loader)

    opt = hub.config.order.apply(raw, raw_cli, cli, cli_args, os_vars, configs)
    # Sync pop-config with the log-level
    opt["pop_config"] = {
        k: opt[sources[0]][k] for k in opt["pop_config"] if k in opt[sources[0]]
    }

    # Output a config based on the current os and cli parameters
    if cli_args.get("config_template"):
        hub.config.template.run(opt)
    if cli_args.get("versions_report"):
        hub.config.versions_report.run(opt, cli)

    hub.OPT = hub.pop.data.imap(opt)

    if logs:
        hub.log[hub.OPT.pop_config.log_plugin].setup(hub.OPT.pop_config)



import locale
import os
import platform
import subprocess
import sys

try:
    from pip._internal.operations.freeze import freeze

    HAS_PIP = True
except ImportError:
    HAS_PIP = False


def run(hub, opts: dict, primary: str):
    """
    Output a config file based on the current command
    :param hub:
    :param opts: The parsed config from os vars, cli options, and config file
    :param primary: The plugin with the authoritative cli

    Returns:
        argv: The exact command ran from the cli
        config: Everything on hub.OPT (which comes from relevant os vars, cli options, and config options)
        freeze: The output of pip freeze
        name: The name of the project with the authoritative cli
        version: The version of the project with the authoritative cli
        python: The python version
        logs: The contents of the most recent log file
    """
    hub.pop.sub.add(dyne_name="rend")
    outputter = opts.get("rend", {}).get("output") or "yaml"

    if HAS_PIP:
        # Get the freeze information from pip internals
        pip_freeze = [line for line in freeze()]
    else:
        # Shell out to pip
        ret = subprocess.check_output([sys.executable, "-m", "pip", "freeze"])
        pip_freeze = ret.decode("utf-8").strip().split("\n")

    log_file = opts.get("pop_config", {}).get("log_file")
    if log_file and os.path.exists(log_file):
        with open(log_file) as fh:
            contents = fh.readlines()
            logs = [c.strip() for c in contents]
    else:
        logs = []

    output = {
        "name": primary,
        "version": hub.config.version.get(primary),
        "argv": " ".join(a for a in sys.argv if a != "--versions-report"),
        "config": hub.config.template.get(opts),
        "python": sys.version,
        "freeze": pip_freeze,
        "locale": locale.getlocale(),
        "platform": {
            "os": os.name,
            "machine": platform.machine(),
            "system": platform.system(),
            "release": platform.release(),
            "version": platform.version(),
        },
        "logs": logs,
    }

    out = hub.output[outputter].display(output)

    print(out)
    sys.exit(0)



import argparse
from typing import Any
from typing import Dict
from typing import List

__func_alias__ = {"help_": "help"}


def init(hub):
    if hub.args.PARSER is None:
        hub.args.PARSER = argparse.ArgumentParser(conflict_handler="resolve")


def help_(hub) -> str:
    """
    Return the help text as a string
    """
    hub.args.parser.init()
    if hub.SUBPARSER:
        return hub.args.SUBPARSERS[hub.SUBPARSER].format_help()
    else:
        return hub.args.PARSER.format_help()


def context(hub) -> Dict[str, Dict[str, Any]]:
    return dict(existing_groups={}, existing_mutex_groups={}, subparser_contexts={})


def apply(
    hub,
    parser: argparse.ArgumentParser,
    arg_definitions: List[Dict[str, Any]],
    context: Dict[str, Any] = None,
):
    """
    :param hub:
    :param parser: An argument parser or subparser
    :param arg_definitions: A dictionary describing an argument
    :param context: values that should be tracked through recursive calls to this function
    """
    if context is None:
        context = hub.args.parser.context()

    for definition in arg_definitions:
        if definition.get("subcommands"):
            hub.args.subcommands.apply(
                root_parser=parser,
                definition=definition,
                context=context,
            )
        elif definition.get("ex_groups"):
            hub.args.parser.apply_ex_groups(parser, context, definition)
        elif definition.get("groups"):
            hub.args.parser.apply_groups(parser, context, definition)
        else:
            hub.args.parser.apply_default(parser, context, definition)


def apply_groups(hub, parser, context, definition):
    for g in definition["groups"]:
        if g not in context["existing_groups"]:
            context["existing_groups"][g] = parser.add_argument_group(g)
        context["existing_groups"][g].add_argument(
            *definition["parser_args"], **definition["parser_kwargs"]
        )


def apply_ex_groups(hub, parser, context, definition):
    for g in definition["ex_groups"]:
        if g not in context["existing_mutex_groups"]:
            context["existing_mutex_groups"][g] = parser.add_mutually_exclusive_group()
        context["existing_mutex_groups"][g].add_argument(
            *definition["parser_args"], **definition["parser_kwargs"]
        )


def apply_default(hub, parser, context, definition):
    parser.add_argument(*definition["parser_args"], **definition["parser_kwargs"])



"""
Define the JSON loader interface
"""
import json

__virtualname__ = "json"


def __virtual__(hub):
    return (True,)


def load(hub, path):
    """
    Use json to read in a file
    """
    try:
        with open(path) as fp_:
            ret = json.loads(fp_.read())
        return ret
    except FileNotFoundError:
        return {}


def render(hub, val):
    """
    Take the string and render it in json
    """
    return json.loads(val)



"""
Support embedding version number lookup into cli
"""
import importlib
import sys


def run(hub, primary):
    """
    Check the version number and then exit
    """
    v = hub.config.version.get(primary)

    print(f"{primary} {v}")
    sys.exit(0)


def get(hub, primary):
    mod = importlib.import_module(f"{primary}.version")
    return mod.version



"""
Gather subcommands for a parser
"""
from typing import Any
from typing import Dict


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    comps = raw_cli[arg]

    subcommands = comps.get("subcommands", [])
    if not isinstance(subcommands, list):
        subcommands = [subcommands]

    return {"subcommands": subcommands}



import logging.handlers


def setup(hub, conf):
    """
    Given the configuration data set up the logger.

    Output to a udp network socket.  If no port is specified, a Unix domain socket is created using the value in `host`

    To specify the host and port add `--log-file="host:port"` to your command line

    I.E.  `--log-file="127.0.0.1:80`

    """
    root = logging.getLogger()

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)

    ff = logging.Formatter(fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"])
    if ":" in conf["log_file"]:
        host, port = conf["log_file"].split(":")
        if port:
            port = int(port)
    else:
        # Open a unix socket
        host = conf["log_file"]
        port = None
    # _, kwargs = hub.render.cli.args(conf["log_handler_options"])
    fh = logging.handlers.DatagramHandler(host=host, port=port)
    fh.setLevel(hub.log.INT_LEVEL)
    fh.setFormatter(ff)
    root.addHandler(fh)



import logging.handlers
import os


def __virtual__(hub):
    return os.name == "nt", "This module is only supported on windows"


def setup(hub, conf):
    """
    Given the configuration data set up the logger.

    Output to a Windows event log
    """
    root = logging.getLogger()

    raw_level = conf["log_level"].strip().lower()
    if raw_level.isdigit():
        hub.log.INT_LEVEL = int(raw_level)
    else:
        hub.log.INT_LEVEL = hub.log.LEVEL.get(raw_level, root.level)

    root.setLevel(hub.log.INT_LEVEL)
    cf = logging.Formatter(fmt=conf["log_fmt_console"], datefmt=conf["log_datefmt"])
    ch = logging.StreamHandler()
    ch.setLevel(hub.log.INT_LEVEL)
    ch.setFormatter(cf)
    root.addHandler(ch)

    ff = logging.Formatter(fmt=conf["log_fmt_logfile"], datefmt=conf["log_datefmt"])
    _, kwargs = hub.render.cli.args(conf["log_handler_options"])
    fh = logging.handlers.NTEventLogHandler(
        conf["log_file"], logtype="Application", **kwargs
    )
    fh.setLevel(hub.log.INT_LEVEL)
    fh.setFormatter(ff)
    root.addHandler(fh)



"""
Gather mutually exclusive groups for a parser or subparser
"""
from typing import Any
from typing import Dict


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    comps = raw_cli[arg]

    if "ex_group" not in comps:
        return {}

    group = comps["ex_group"]

    return {"ex_groups": [group]}



from typing import Any
from typing import Dict

import dict_tools.update as dict_tool


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    raw_cli.get(arg, {}).copy()

    result = {
        "parser_kwargs": {},
        "parser_args": [],
        "groups": {},
        "ex_groups": {},
        "subcommands": [],
    }

    for sub in hub.comps._loaded:
        if sub == "init":
            continue
        ret = hub.comps[sub].get(raw_cli, arg)
        result = dict_tool.update(result, ret, merge_lists=True)

    return result



import sys
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import dict_tools.update


def get(hub, raw: Dict[str, Any], cli: str) -> Dict[str, Any]:
    """
    Gather the arguments that need to be parsed by the CLI
    """
    ret = {}
    main = raw.get(cli, {}).get("CLI_CONFIG")
    main_raw = raw.get(cli, {}).get("CONFIG")

    for key, data in main.items():
        ret[key] = {}
        dict_tools.update.update(ret[key], data)
        if key in main_raw:
            dict_tools.update.update(ret[key], main_raw[key])
        if "source" in data:
            src = raw.get(data["source"], {}).get("CONFIG", {}).get(key)
            if src is not None:
                dict_tools.update.update(ret[key], src)
        if "default" in ret[key]:
            ret[key]["default"] = hub.args.DEFAULT
    return ret


def gather(
    hub, raw: Dict[str, Any], cli: str, parse_cli: bool
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Return the cli arguments as they are parsed
    """
    if not parse_cli:
        return {}, {}
    hub.args.parser.init()
    hub.args.subcommands.parse(raw, cli)

    raw_cli = hub.args.cli.get(raw, cli)
    hub.args.init.setup(raw_cli)

    cli_args = hub.args.cli.parse(sys.argv[1:])
    hub.args.cli.render(cli_args, raw_cli)

    clean_args = hub.args.cli.clean_defaults(cli_args)
    return clean_args, raw_cli


def parse(
    hub,
    args: List[str],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Parse the command line options
    """
    opts = hub.args.PARSER.parse_args(args)
    opts_dict = opts.__dict__
    hub.SUBPARSER = opts_dict.get("_subparser_", None)

    # Parse the args of the subparser
    if hub.SUBPARSER:
        # Remove the subparser and re-parse what's left so that we can pick up anything that was missed
        args.remove(hub.SUBPARSER)
        sub_opts, _ = hub.args.SUBPARSERS[hub.SUBPARSER].parse_known_args(args)
        sub_opts_dict = sub_opts.__dict__
    else:
        sub_opts_dict = {}

    # If the subparser got any values that should have been passed to the root, then update the dict
    for k in opts_dict:
        if opts_dict[k] is hub.args.DEFAULT:
            sub_opt = sub_opts_dict.get(k, hub.args.DEFAULT)
            if sub_opt is not hub.args.DEFAULT:
                opts_dict[k] = sub_opt
    return opts_dict


def clean_defaults(hub, cli_args: Dict[str, Any]) -> Dict[str, Any]:
    """
    If anyone did not pass in an argument then the key will match the
    bad default and needs to be removed
    """
    ret = {}
    for key, val in cli_args.items():
        if val is not hub.args.DEFAULT:
            ret[key] = val
    return ret


def render(hub, cli_args: Dict[str, Any], raw_cli: Dict[str, Any]):
    """
    For options specified as such, take the string passed into the cli and
    render it using the specified render flag
    """
    for key, val in raw_cli.items():
        if key not in cli_args:
            continue
        if "render" not in val:
            continue
        if val["default"] != cli_args[key]:
            # The value was changed, render it
            cli_args[key] = hub.render.init.process(val["render"], cli_args[key])



"""
Gather groups for a parser or subparser
"""
from typing import Any
from typing import Dict


def get(hub, raw_cli: Dict[str, Any], arg: str) -> Dict[str, Any]:
    comps = raw_cli[arg]

    if "group" not in comps:
        return {}

    group = comps["group"]

    return {"groups": [group]}



import os
import pathlib


class _DefaultOption:
    pass


def _insert_default_placeholders(raw):
    defaults = {}
    for imp in raw:
        defaults[imp] = {}
        for key, data in raw[imp]["CONFIG"].items():
            if "default" not in data:
                raise KeyError(f"No default value for '{key}' in '{imp}'s conf.py")
            defaults[imp][key] = data["default"]
            data["default"] = _DefaultOption
    return defaults


def _restore_raw_defaults(raw, defaults):
    # undo our modification of the raw data structure
    for imp in raw:
        for key, data in raw[imp]["CONFIG"].items():
            raw[imp]["CONFIG"][key]["default"] = defaults[imp][key]


def _replace_default_placeholders(ret, defaults):
    for imp in ret:
        for key, data in ret[imp].items():
            if data is _DefaultOption:
                ret[imp][key] = defaults[imp][key]
    return ret


def _reroot_paths(defaults, root):
    for imp in defaults:
        for key, val in defaults[imp].items():
            try:
                path = pathlib.Path(val or "")
            except TypeError:
                continue
            if key == "root_dir":
                defaults[imp][key] = root
            elif (key.endswith(("_dir", "_path", "_file"))) and path.is_absolute():
                # only update absolute paths for keys
                #  ending in _dir, _path or _file
                defaults[imp][key] = _reroot_path(val, imp, root)


def _get_root(ret, cli):
    if "root_dir" not in ret.get(cli, {}):
        # there is no root_dir parameter, do not activate roots system
        #  otherwise there would be no way to *disable* the system
        #  either by leaving out root_dir, or manually specifying root_dir=/
        return None

    # root_dir is not configured, maybe use home directory
    if ret[cli]["root_dir"] is _DefaultOption:
        if hasattr(os, "geteuid") and os.geteuid() != 0:
            # Using pathlib.Path on all paths makes testing on multiple OSes much easier.
            # Since we can just patch pathlib.Path to be pathlib.PureWindowsPath or pathlib.PurePosixPath
            return str(pathlib.Path(os.path.expanduser("~")) / f".{cli}")
        else:
            return None
    else:
        return ret[cli]["root_dir"]


def _reroot_path(val, imp, new_root):
    root = pathlib.Path(new_root)
    path = pathlib.Path(val)
    if path.is_absolute() and imp in path.parts:
        # remove the root from the path, then join it to the main root
        return str(root.joinpath(*path.parts[1:]))
    return val


def pre_apply(hub, ctx):
    kwargs = ctx.get_arguments()
    ctx.cache["root_defaults"] = _insert_default_placeholders(kwargs["raw"])


def post_apply(hub, ctx):
    kwargs = ctx.get_arguments()
    _restore_raw_defaults(kwargs["raw"], ctx.cache["root_defaults"])
    root = _get_root(ctx.ret, kwargs["cli"])
    if root:
        _reroot_paths(ctx.cache["root_defaults"], root)
    _replace_default_placeholders(ctx.ret, ctx.cache["root_defaults"])
